from __future__ import print_function
import sys
import math
import numpy as np 

VERSION='20231001'

list_telescope=['Araki','NTT','Magellan','Keck']
list_mode=['WIDE','HIRES-Y','HIRES-J']
list_slit=['100','140','200']
list_season=['winter','summer']
list_nod_pattern=['AB', 'ABBA', 'OS', 'OSO', 'ABO']
list_overhead={'acquisition':300., 'readout':15., 'nodding':15.}
list_wranges={'WIDE':[9100.,13500.],'HIRES-Y':[9550.,11100.],'HIRES-J':[11350.,13500.]}
eff_array=1.0
eff_telluric=0.95

min_itime=2.
max_itime_default=900.
default_frame_itime=300.

arg_helps={
        'telescope':'Telescope',
        'mode':'Mode',
        'slit':'Slit width in micron',
        'season':'Season (winter or summer)',
        'seeing':'Seeing(FWHM) in arcsec',
        'frame_itime':'Integraion time per frame in sec',
        'nframe':'Number of frames (including sky when OSO)',
        'nod_pattern':'Dithering Patern (ABBA or OSO)',
        'cor_factor':'Correction Factor' ,
        'SN':'Required S/N' }

arg_defaults={
        'telescope':'Magellan',
        'mode':'WIDE',
        'slit':'100',
        'season':'winter',
        'nyquist':False,
        'seeing':0.8,
        'psf_size':0.8,
        'airmass':1.0,
        'SN':np.nan,
        'frame_itime':default_frame_itime,
        'nframe':np.nan,
        'nod_pattern':'ABBA',
        'cor_factor':1.0,
        'ymag':np.nan,
        'jmag':np.nan,
        'spectrum':False
        }

option_short={
        'verbose':'-v', 'nyquist':'-n', 'cor_factor':'-c', 'telescope':'-t', 'spectrum':'-s', 'max_itime_total':'-xt', 'max_itime':'-xi', 'jmag':'-j', 'ymag':'-y', 'header':'-hd', 'aj':'-aj', 'ak':'-ak' }

option_helps={
        'verbose':'Verbose',
        'nyquist':'S/N with Nyquist sampling [per half resolution] (otherwise pixel sampling)',
        'cor_factor':'Arbitrary correction factor of sensitivity',
        'telescope':'select the telescope (Araki, NTT, Magellan, Keck; default=Magellan)',
        'jmag':'J-band magnitude (mag)',
        'ymag':'Y-band magnitude (mag)',
        'spectrum':'Give the spectrum to stdin (wave in AA, flux in W/m^3)',
        'max_itime_total':'Maximum total integration time (sec)',
        'max_itime':'Maximum integration time per frame (sec)',
        'header':'Print out the csv header',
        'step':'Wavelength step (AA)',
        'teff':'Effective temperature (K)',
        'aj':'J-band extinction (mag)',
        'ak':'K-band extinction (mag)'
        }

option_defaults={ 'verbose': False, 'nyquist': False, 'telescope':'Magellan', 'cor_factor':1.0, 'max_itime_total':36000., 'max_itime':max_itime_default, 'ymag':np.nan, 'jmag':np.nan, 'spectrum':False, 'header':False, 'step':np.nan, 'teff':np.nan, 'aj':np.nan, 'ak':np.nan, 'name':None }

readnoise=6.1

num_nodding_max=50

def get_resolution(mode, slit):
    if mode=='WIDE':
        resolution={'100':28000.,'140':24000.,'200':18000.}[str(slit)]
        pixel_resolution=56000.
    elif (mode=='HIRES-Y') or (mode=='HIRES-J'):
        resolution={'100':70000.,'140':58000.,'200':44000.}[str(slit)]
        pixel_resolution=160000.
    return resolution, pixel_resolution

telescope_params={
    'Araki':{
        'observatory':'Koyama Astronomical Observatory',
        'reference':'http://www.kyoto-su.ac.jp/observatory/shisetu.html',
        'diameter':1.3,
        'eff_mirrors':0.73,
        'eff_f_ratio':0.97,
        'eff_secondary':0.9,
        'pixel_scale':0.8,
        'default_seeing':3.0,
        'th_background_summer':0.15,
        'th_background_winter':0.07
        },
    'NTT':{
        'observatory':'La Silla Observatory, ESO',
        'reference':'https://www.eso.org/sci/facilities/lasilla/telescopes/ntt/overview.html',
        'diameter':3.58,
        'eff_mirrors':0.73,
        'eff_f_ratio':1.0,
        'eff_secondary':0.895,
        'pixel_scale':0.27,
        'default_seeing':1.4,
        'th_background_summer':0.10,
        'th_background_winter':0.10
        },
    'Magellan':{
        'observatory':'Las Campanas Observatory',
        'reference':'http://www.lco.cl/telescopes-information/magellan/',
        'diameter':6.5,
        'eff_mirrors':0.73,
        'eff_f_ratio':0.97,
        'eff_secondary':0.926,
        'pixel_scale':0.15,
        'default_seeing':0.8,
        'th_background_summer':0.12,
        'th_background_winter':0.08
        },
    'Keck':{
        'observatory':'W.M. Keck Observatory',
        'reference':'http://www.keckobservatory.org/',
        'diameter':9.96,
        'eff_mirrors':0.73,
        'eff_f_ratio':0.95,
        'eff_secondary':1.0,
        'pixel_scale':0.092,
        'default_seeing':0.6,
        'th_background_summer':0.07,
        'th_background_winter':0.07
        }
    }

def count_nod_numbers(nod_sequence, verbose=False):
    ntarget, nsky, n_nod = 0, 0, 0
    for i, pos in enumerate(list(nod_sequence)):
        if pos == 'S':
            nsky += 1
        else:
            ntarget += 1
        if (i!=0) and (pos!=nod_sequence[i-1]):
            n_nod += 1
    if verbose is True:
        print('   nod_sequence={0} -> ntarget={1:d}, nsky={2:d}, n_nod={3:d}'.format(nod_sequence,ntarget,nsky,n_nod),file=sys.stderr)
    return ntarget, nsky, n_nod

def make_nod_sequence(nod_pattern, nframe, verbose=False):
    if nod_pattern not in list_nod_pattern:
        return 
    ntarget, nsky=0, 0
    tmp_sequence = ''
    while len(tmp_sequence)<nframe:
        tmp_sequence+=nod_pattern
    nod_sequence=tmp_sequence[0:nframe]
    if verbose is True:
        print('   nod_pattern={0} -> nod_sequence={1}'.format(nod_pattern,nod_sequence),file=sys.stderr)
    return nod_sequence

def init_setting():
    p={}
    for arg, value in arg_defaults.items():
        p[arg]=value
    for option, value in option_defaults.items():
        p[option]=value
    return p

def validate_setting(setting, tel_params):
    if setting['telescope'] not in list_telescope:
        return False
    if setting['mode'] not in list_mode:
        return False
    if setting['slit'] not in list_slit:
        return False
    if setting['season'] not in list_season:
        return False
    if setting['nod_pattern'] not in list_nod_pattern:
        return False
    if setting['seeing']<=0:
        return False
    #if (not np.isnan(setting['frame_itime'])) and ((setting['frame_itime']<=min_itime) or (setting['frame_itime']>setting['max_itime'])):
    if (not np.isnan(setting['frame_itime'])) and (setting['frame_itime']<=min_itime):
        return False
    if (not np.isnan(setting['nframe'])) and (setting['nframe']<=0):
        return False
    if (not np.isnan(setting['SN'])) and (setting['SN']<=0):
        return False
    if setting['cor_factor']<=0:
        return False
    if setting['verbose'] is True:
        print_setting(setting)
    setting['efficiency_others']=calc_efficiency(setting, tel_params)
    setting['psf_size']=setting['seeing']
    aperture_space, aperture_wave=get_aperture_values(setting, tel_params)
    setting['aperture_space']=aperture_space
    setting['aperture_wave']=aperture_wave
    resolution, pixel_resolution=get_resolution(setting['mode'],setting['slit'])
    setting['resolution']=resolution
    setting['pixel_resolution']=pixel_resolution
    if setting['nyquist'] is True:
        setting['unit_resolution']=2.*resolution
    else:
        setting['unit_resolution']=pixel_resolution
    setting['num_integ_pixel']=calc_num_integ_pixel(setting,tel_params)
    setting['slit_integ_fraction']=calc_slit_integ_fraction(setting)
    return True

def print_setting(p):
    print('# Input parameters:',file=sys.stderr)
    print('   telescope={0}  mode={1}  slit={2}'.format(p['telescope'],p['mode'],p['slit']),file=sys.stderr)
    print('   season={0}  seeing={1:.2f}(arcsec)'.format(p['season'],p['seeing']),file=sys.stderr)
    if p['nyquist'] is True:
        print('   Nyquist sampling (per half resolution)',file=sys.stderr)
    else:
        print('   Non-Nyquist sampling (per pixel)',file=sys.stderr)
    if not np.isnan(p['ymag']):
        print('   ymag={0:.3f}(mag)'.format(p['ymag']),file=sys.stderr)
    if not np.isnan(p['jmag']):
        print('   jmag={0:.3f}(mag)'.format(p['jmag']),file=sys.stderr)
    if p['spectrum'] is True:
        print('   spectrum input should be given',file=sys.stderr)
    for key in ['frame_itime','nframe','SN']:
        if not np.isnan(p[key]):
            print('   {0}={1}'.format(key,str(p[key])),file=sys.stderr)

def get_aperture_values(setting, tel_params, aperture_space_in_fwhm=2.):
    pixel_scale=tel_params['pixel_scale']
    if 'psf_size' in setting:
        aperture_space=setting['psf_size']*aperture_space_in_fwhm
    else:
        aperture_space=setting['seeing']*aperture_space_in_fwhm
    if setting['slit']=='100':
        aperture_wave=pixel_scale*2.0
    elif setting['slit']=='140':
        aperture_wave=pixel_scale*2.8
    elif setting['slit']=='200':
        aperture_wave=pixel_scale*4.0
    return aperture_space, aperture_wave

def get_telescope_params(telescope_name):
    if telescope_name not in telescope_params:
        print('Unexpected telescope_name={0}'.format(telescope_name),file=sys.stderr)
        print('   Available telescopes: {0}'.format(', '.join(list_telescope)),file=sys.stderr)
        return None
    return telescope_params[telescope_name]
            
def get_th_background(tel_params, season):
    return tel_params['th_background_'+season]

def calc_throughput(mode, wave_AA):
    coef={'WIDE-Y':{'a':1.333E-4,'b':-0.967}, 'WIDE-J':{'a':0.0,'b':0.55}, 'HIRES-Y':{'a':1.333E-4,'b':-1.063}, 'HIRES-J':{'a':0.0,'b':0.42}}
    if mode=='WIDE':
        if wave_AA>11250.:
            throughput=coef[mode+'-J']['a']*wave_AA+coef[mode+'-J']['b']
        else:
            throughput=coef[mode+'-Y']['a']*wave_AA+coef[mode+'-Y']['b']
    else:
        throughput=coef[mode]['a']*wave_AA+coef[mode]['b']
    return throughput

def calc_eff_telluric(airmass):
    return np.exp(-0.05*airmass)

def calc_efficiency(setting, tel_params, silence=False):
    efficiency=eff_array
    eff_telluric=calc_eff_telluric(setting['airmass'])
    efficiency*=eff_telluric
    for key in ['eff_mirrors','eff_secondary','eff_f_ratio']:
        efficiency*=tel_params[key]
    efficiency*=setting['cor_factor']
    if setting['verbose'] is True:
        print('# Efficiency calculation:',file=sys.stderr)
        print('   {0}={1:.3f}'.format('eff_array',eff_array),file=sys.stderr)
        print('   {0}={1:.3f}'.format('eff_telluric',eff_telluric),file=sys.stderr)
        for key in ['eff_mirrors','eff_f_ratio','eff_secondary']:
            print('   {0}={1:.3f}'.format(key,tel_params[key]),file=sys.stderr)
        print('   {0}={1:.3f}'.format('cor_factor',setting['cor_factor']),file=sys.stderr)
        print('   -> efficiency without throughput={0:.3f}'.format(efficiency),file=sys.stderr)
    return efficiency

def calc_num_integ_pixel(setting, tel_params):
    if setting['nyquist'] is True:
        num_integ_pixel=(setting['aperture_space']/tel_params['pixel_scale'])*(setting['pixel_resolution']/setting['unit_resolution'])
        if setting['verbose'] is True:
            print('# Pixel number for sampling unit (Nyquist sampling)',file=sys.stderr)
            print('   (aperture_space={0:.3f}sec) / (pixel_scale={1:.3f}sec/pix) = {2:.3f} pix per column'.format(setting['aperture_space'],tel_params['pixel_scale'],setting['aperture_space']/tel_params['pixel_scale']),file=sys.stderr)
            print('   (pixel_resolution={0:.3f}sec) / (unit_resolution={1:.3f}sec/pix) = {2:.3f} pix per row'.format(setting['pixel_resolution'],setting['unit_resolution'],setting['pixel_resolution']/setting['unit_resolution']),file=sys.stderr)
            print('   -> num_integ_pixel = {0:.3f}pix'.format(num_integ_pixel),file=sys.stderr)
    else:
        num_integ_pixel=setting['aperture_space']/tel_params['pixel_scale']
        if setting['verbose'] is True:
            print('# Pixel number for sampling unit (S/N per spectral pixel)',file=sys.stderr)
            print('   (aperture_space={0:.3f}) / (pixel_scale={1:.3f}sec/pix) = {2:.3f}pix'.format(setting['aperture_space'],tel_params['pixel_scale'],num_integ_pixel),file=sys.stderr)
    return num_integ_pixel


def calc_slit_integ_fraction(setting):
    psf_size=setting['psf_size']
    aperture_space=setting['aperture_space']
    aperture_wave=setting['aperture_wave']
    sigma=psf_size/2.354820
    sigma_space_max=aperture_space/sigma/2.0
    sigma_wave_max=aperture_wave/sigma/2.0
    space_fraction, wave_fraction=0., 0.
    sigma_pos, delta_sigma=0., 0.01
    while sigma_pos<sigma_space_max:
        space_fraction+=(delta_sigma/np.sqrt(2.0*np.pi))*math.exp(-0.5*(sigma_pos+delta_sigma/2.)**2)
        sigma_pos+=delta_sigma
    space_fraction*=2.
    sigma_pos, delta_sigma=0., 0.01
    while sigma_pos<sigma_wave_max:
        wave_fraction+=(delta_sigma/np.sqrt(2.0*np.pi))*math.exp(-0.5*(sigma_pos+delta_sigma/2.)**2)
        sigma_pos+=delta_sigma
    wave_fraction*=2.
    slit_integ_fraction=space_fraction*wave_fraction
    if setting['verbose'] is True:
        print('# Slit throughput calculation:',file=sys.stderr)
        print('   psf_size={0:.2f}(arcsec)  sigma={1:.2f}(arcsec)'.format(psf_size,sigma),file=sys.stderr)
        print('   aperture_space ={0:.2f}(arcsec) = +- {1:.2f}sigma -> space_fraction = {2:.3f}'.format(aperture_space,sigma_space_max,space_fraction),file=sys.stderr)
        print('   aperture_wave  ={0:.2f}(arcsec) = +- {1:.2f}sigma -> wave_fraction  = {2:.3f}'.format(aperture_wave,sigma_wave_max,wave_fraction),file=sys.stderr)
        print('   -> slit_integ_fraction = {0:.3f}'.format(slit_integ_fraction),file=sys.stderr)
    return slit_integ_fraction
