from __future__ import print_function
import sys
from math import ceil
import numpy as np
import params

const_h=6.626068E-34
const_c=2.99792458E+08
const_k=1.38E-23


def calc_signal(setting, tel_params, frame_itime, nod_sequence, spec, silence=False):
    ntarget, nsky, n_nod=params.count_nod_numbers(nod_sequence,verbose=False)
    diameter=tel_params['diameter'] 
    area=np.pi*(tel_params['diameter']/2)**2
    total_itime=frame_itime*ntarget
    list_signal=[]
    for i, (wave_AA, flux) in enumerate(spec):
        wave=wave_AA*1.0E-10
        delta_wave=wave/setting['unit_resolution']
        photon_per_energy=wave/(const_h*const_c)
        throughput=params.calc_throughput(setting['mode'],wave_AA)
        signal=throughput*setting['efficiency_others']*setting['slit_integ_fraction']*flux*delta_wave*photon_per_energy*area*total_itime
        list_signal.append(signal)
        if (silence is False) and (setting['verbose'] is True) and (i==0):
            print('# Signal ({0:d} target frames):'.format(ntarget), file=sys.stderr)
            print('   wave={0:.3e}(m)  flux={1:.3e}(W/m^3)'.format(wave,flux), file=sys.stderr)
            print('   total efficiency={0:.3f} (throughput={1:.3f})'.format(throughput*setting['efficiency_others'],throughput), file=sys.stderr)
            print('   photon_per_energy={0:.3e}(photon/J)'.format(photon_per_energy), file=sys.stderr)
            print('   delta_wave={0:.3e}(m per sampling)  (Runit={1:.0f})'.format(delta_wave,setting['unit_resolution']), file=sys.stderr)
            print('   area=pi*({0:.2f}/2)^2={1:.3f}(m^2)'.format(tel_params['diameter'],area), file=sys.stderr)
            print('   frame_itime={0:.1f} * ntarget={1:d} -> itime={2:.1f}'.format(frame_itime,ntarget,total_itime), file=sys.stderr)
            print('   efficiency={0:.4f}   slit_integ_fraction={1:.4f}'.format(setting['efficiency_others']*throughput,setting['slit_integ_fraction']), file=sys.stderr)
            print('   -> signal={0:.0f}(e- per sampling)'.format(signal), file=sys.stderr)
    return list_signal

def calc_noise(setting, tel_params, frame_itime, nod_sequence, list_signal, silence=False):
    ntarget, nsky, n_nod=params.count_nod_numbers(nod_sequence,verbose=False)
    npix=setting['num_integ_pixel']
    th_background=params.get_th_background(tel_params,setting['season'])
    readnoise=params.readnoise
    bg_level=th_background*frame_itime*npix
    noise_read=np.sqrt(2*ntarget*npix)*readnoise
    noise_bg=np.sqrt(2*ntarget*bg_level)
    list_noise=[]
    for i, signal in enumerate(list_signal):
        noise_signal=np.sqrt(signal)
        noise_total=np.sqrt(noise_signal**2+noise_read**2+noise_bg**2)
        if (silence is False) and (setting['verbose'] is True) and (i==0):
            print('# Noise calculation (ntarget={0:d}, nsky={1:d})'.format(ntarget,nsky),file=sys.stderr)
            print('   num_integ_pixel={0:.2f}'.format(npix),file=sys.stderr)
            print('   signal={0:.0f}(e- per sampling)  sqrt(signal)={1:.1f}(e- per sampling)'.format(signal,noise_signal),file=sys.stderr)
            print('   th_background={0:.3f}(e-/s/pix)  (season={1})'.format(th_background,setting['season']),file=sys.stderr)
            print('   background = {0:.3f}(e-/s/pix) * {1:.2f}(s) * {2:.2f}(pix) = {3:.2f} (e- per sampling)'.format(th_background,frame_itime,npix,bg_level),file=sys.stderr)
            print('   sqrt((2*{0:d})*background)={1:.2f} (e- per sampling)'.format(ntarget,noise_bg),file=sys.stderr)
            print('   readnoise={0:.2f}(e-/pix) * sqrt((2*{1:d})*{2:.2f}pix) = {3:.2f}(e-)'.format(readnoise,ntarget,npix,noise_read),file=sys.stderr)
            print('   -> noise={0:.2f}(e- per sampling)'.format(noise_total),file=sys.stderr)
        list_noise.append([noise_total, {'noise_signal':noise_signal, 'noise_bg':noise_bg, 'noise_read':noise_read}])
    return list_noise

def select_dominant_noise(noises):
    if (noises['noise_signal']>=noises['noise_bg']) and (noises['noise_signal']>=noises['noise_read']):
        return 'noise_signal'
    if (noises['noise_bg']>=noises['noise_signal']) and (noises['noise_bg']>=noises['noise_read']):
        return 'noise_bg'
    if (noises['noise_read']>=noises['noise_signal']) and (noises['noise_read']>=noises['noise_bg']):
        return 'noise_read'
    return None

def calc_SN(setting,tel_params,frame_itime,nod_sequence,spec,silence=False):
    list_signal=calc_signal(setting, tel_params, frame_itime, nod_sequence, spec, silence=silence)
    list_noise=calc_noise(setting, tel_params, frame_itime, nod_sequence, list_signal, silence=silence)
    list_SN=[]
    for i, (signal, (noise_total, noises)) in enumerate(zip(list_signal,list_noise)):
        SN=signal/noise_total
        list_SN.append(SN)
        if (silence is False) and (setting['verbose'] is True) and (i==0):
            dominant_noise=select_dominant_noise(noises)
            if dominant_noise=='noise_signal':
                print('# Poisson noise is dominant',file=sys.stderr)
            if dominant_noise=='noise_bg':
                print('# Thermal background noise is dominant',file=sys.stderr)
            if dominant_noise=='noise_read':
                print('# Readout noise is dominant',file=sys.stderr)
            print('   '+', '.join(['{0}={1:.2f}e-'.format(s,n) for s, n in noises.items()]),file=sys.stderr)
            print('# S/N calculation',file=sys.stderr)
            print('   S/N = (signal={0:.2f}e-) / (noise={1:.2f}e-) = {2:.1f}'.format(signal,noise_total,SN),file=sys.stderr)
    return list_SN

def search_itime_and_sequence(setting,tel_params,wave,flux,requestSN,silence=False,max_itime=params.max_itime_default):
    if (silence is not True) and (setting['verbose'] is True):
        print('# search_itime_and_sequence',file=sys.stderr)
    spec=[[wave,flux]]
    min_itime=params.min_itime
    max_itime_total=setting['max_itime_total']
    nod_pattern=setting['nod_pattern']
    ### check the minimum SN achieved with min_itime and 2 frames
    if 'S' in nod_pattern:
        nod_sequence='OS'
    else:
        nod_sequence='AB'
    min_SN=(calc_SN(setting,tel_params,min_itime,nod_sequence,spec,silence=True))[0]
    if (silence is not True) and (setting['verbose'] is True):
        print('   {0:.0f} * {1} (total={2:.0f}) -> min_S/N={3:.2f}'.format(min_itime,nod_sequence,min_itime*len(nod_sequence),min_SN),file=sys.stderr)
    if min_SN>=requestSN:
        return min_itime, nod_sequence
    ### search for N*nod_pattern that satisfies requestSN
    for n in np.arange(1,30):
        nod_sequence=n*nod_pattern
        ntarget, nsky, n_nod=params.count_nod_numbers(nod_sequence,verbose=False)
        total_itime=(ntarget+nsky)*max_itime
        SN=(calc_SN(setting,tel_params,max_itime,nod_sequence,spec,silence=True))[0]
        if SN>=requestSN:
            if (silence is not True) and (setting['verbose'] is True):
                print('   {0:.0f}(=max) * {1} (total={2:.0f}) -> S/N={3:.2f}'.format(max_itime,nod_sequence,total_itime,SN),file=sys.stderr)
            break
        elif total_itime+(len(nod_pattern)*max_itime)>max_itime_total:
            if (silence is not True) and (setting['verbose'] is True):
                print('   {0:.0f} * {1} (total={2:.0f}=max) -> max_S/N={3:.2f}'.format(max_itime,nod_sequence,total_itime,SN),file=sys.stderr)
            return max_itime, nod_sequence
    ### serach for frame_itime that satisfies requestSN with a given nod_sequence
    itime0, itime1=min_itime, max_itime
    SN0=(calc_SN(setting,tel_params,itime0,nod_pattern,spec,silence=True))[0]
    if SN0>=requestSN:
        if (silence is not True) and (setting['verbose'] is True):
            print('   {0:.0f} * {1}(=min) (total={2:.0f}) -> S/N={3:.2f}'.format(itime0,nod_sequence,itime0*len(nod_sequence),SN0),file=sys.stderr)
        return itime0, nod_sequence
    #SN1=(calc_SN(setting,tel_params,itime1,nod_pattern,spec,silence=True))[0]
    while itime1-itime0>0.2:
        itime_new=(itime0+itime1)/2.
        SN_new=(calc_SN(setting,tel_params,itime_new,nod_sequence,spec,silence=True))[0]
        if SN_new<requestSN:
            itime0=itime_new
        else:
            itime1=itime_new
    frame_itime=ceil((itime0+itime1)/2.)
    SN=(calc_SN(setting,tel_params,frame_itime,nod_pattern,spec,silence=True))[0]
    if (silence is not True) and (setting['verbose'] is True):
        print('   {0:.0f} * {1} (total={2:.0f}) -> S/N={3:.2f}'.format(frame_itime,nod_sequence,frame_itime*len(nod_sequence),SN),file=sys.stderr)
    return frame_itime, nod_sequence

def calc_overhead(nframe,n_nod,verbose=False):
    overhead_aq=params.list_overhead['acquisition']
    overhead_read=params.list_overhead['readout']*nframe
    overhead_nod=params.list_overhead['nodding']*n_nod
    overhead=overhead_aq+overhead_read+overhead_nod
    if verbose is True:
        print('# Overhead calculation',file=sys.stderr)
        print('   overhead={0:.0f}s (acquisition={1:.0f}s, readout={2:.0f}s*{3:d}, nodding={4:.0f}s*{5:d})'.format(overhead,overhead_aq,params.list_overhead['readout'],nframe,params.list_overhead['nodding'],n_nod),file=sys.stderr)
    return overhead

def load_template(file_in, wmin, wmax, flux_factor=1.):
    waves=[]; fluxs=[]
    data=np.loadtxt(file_in)
    for row in data:
        wave=row[0]
        if (wave>=wmin) and (wave<=wmax):
            flux=row[1]*flux_factor
            waves.append(wave)
            fluxs.append(flux)
    return np.array(waves), np.array(fluxs)

def calc_BB(Teff, wave_AA, normalize=None):
    if normalize is None:
        wave=wave_AA*1.0E-10
        BB=(2.*const_h*const_c**2/np.power(wave,5))/(np.exp(const_h*const_c/(wave*const_k*Teff))-1.)
    else:
        if ('wave' not in normalize) or ('flux' not in normalize):
            return None
        wave_zero=normalize['wave']
        flux_zero=normalize['flux']
        if (wave_zero<=0.) or (flux_zero<=0.):
            return None
        BB=(flux_zero/calc_BB(Teff,wave_zero,normalize=None))*calc_BB(Teff,wave_AA,normalize=None)
    return BB

def apply_extinction(waves, fluxs, aj):
    fluxs_new=[]
    if np.isnan(aj):
        return fluxs
    for w, f in zip(waves,fluxs):
        ext_factor=pow(10.,-0.4*(aj*(pow(w/12500.,-1.99)-1)))
        fluxs_new.append(f*ext_factor)
    return fluxs_new


